import typing

from blacksheep.messages import Request
from blacksheep.middlewares import MiddlewareCategory
from blacksheep.server.security.hsts import HSTSMiddleware

if typing.TYPE_CHECKING:
    from blacksheep.server.application import Application


class HTTPSchemeMiddleware:
    """
    Middleware that forces request.scheme based on configuration.
    Useful when the application is deployed behind proxies that do TLS termination, and
    the application needs to generate proper redirect URLs to itself with the right
    scheme.
    This middleware is applied automatically when the env variables APP_HTTP_SCHEME or
    APP_FORCE_HTTPS are set.
    """

    def __init__(self, scheme: str):
        if scheme not in {"http", "https"}:
            raise TypeError("Invalid scheme, expected http | https")
        self.scheme = scheme

    async def __call__(self, request: Request, handler):
        request.scheme = self.scheme
        return await handler(request)


def configure_scheme_middleware(app: "Application"):
    """
    Automatically configures request scheme handling based on environment variables.

    This function is useful when the application runs behind reverse proxies or load
    balancers that perform TLS termination, ensuring the application generates URLs
    with the correct scheme.

    Environment Variables:
        APP_FORCE_HTTPS: When set to a truthy value ("1", "true", etc.), forces all
            requests to use https scheme and adds HSTS headers for security.
        APP_HTTP_SCHEME: When set to "http" or "https", forces all requests to use
            the specified scheme without adding HSTS headers.

    Behavior:
        - If APP_FORCE_HTTPS is set: Uses https + enables HSTS middleware
        - If APP_HTTP_SCHEME is set: Uses the specified scheme without HSTS
        - If neither is set: No middleware is applied (uses actual request scheme)

    Note: APP_FORCE_HTTPS takes precedence over APP_HTTP_SCHEME if both are set.
    """
    if app.env_settings.force_https:
        # Apply middleware that configures request.scheme to match env settings
        app.middlewares.append(
            HTTPSchemeMiddleware("https"), MiddlewareCategory.INIT, -100
        )

        # Apply HTTP Strict Transport Security header by default
        app.middlewares.append(HSTSMiddleware(), MiddlewareCategory.MESSAGE)
    elif app.env_settings.http_scheme:
        app.middlewares.append(
            HTTPSchemeMiddleware(app.env_settings.http_scheme),
            MiddlewareCategory.INIT,
            -100,
        )



"""
Common types annotations and functions.
"""

from typing import AnyStr, Iterable

from blacksheep.url import URL

KeyValuePair = tuple[AnyStr, AnyStr]
HeadersType = dict[AnyStr, AnyStr] | Iterable[KeyValuePair]
ParamsType = dict[AnyStr, AnyStr] | Iterable[KeyValuePair]
URLType = str | bytes | URL


def _ensure_header_bytes(value: AnyStr) -> bytes:
    return value if isinstance(value, bytes) else value.encode("ascii")


def _ensure_param_str(value: AnyStr) -> str:
    return value if isinstance(value, str) else value.decode("ascii")


def normalize_headers(
    headers: HeadersType | None,
) -> list[tuple[bytes, bytes]] | None:
    if headers is None:
        return None
    if isinstance(headers, dict):
        return [
            (_ensure_header_bytes(key), _ensure_header_bytes(value))  # type: ignore
            for key, value in headers.items()
        ]
    return [
        (_ensure_header_bytes(key), _ensure_header_bytes(value))
        for key, value in headers
    ]


def normalize_params(params: ParamsType | None) -> list[tuple[str, str]] | None:
    if params is None:
        return None
    if isinstance(params, dict):
        return [
            (_ensure_param_str(key), _ensure_param_str(value))  # type: ignore
            for key, value in params.items()
        ]
    return [(_ensure_param_str(key), _ensure_param_str(value)) for key, value in params]



"""
This module provides classes to handle Basic Authentication.
"""

import base64
import secrets
from abc import ABC, abstractmethod
from copy import deepcopy

from essentials.secrets import Secret
from guardpost import AuthenticationHandler, Identity
from guardpost.errors import InvalidCredentialsError

from blacksheep.messages import Request


class BasicCredentials:
    """
    Represents a set of Basic Authentication credentials with username and password.
    """

    def __init__(
        self,
        username: str,
        password: Secret,
        claims: dict | None = None,
        roles: list[str] | None = None,
    ) -> None:
        """
        Initialize Basic Authentication credentials.

        Parameters
        ----------
        username : str
            The username for authentication.
        password : Secret
            The password for authentication, stored securely.
        claims : dict | None, optional
            Additional claims to include in the authenticated identity, by default None.
        roles : list[str] | None, optional
            List of roles to assign to the authenticated identity, by default None.
        """
        self._username = username
        self._password = password
        self._claims = claims or {}
        self._roles = roles or []

    @property
    def username(self) -> str:
        """Returns the username."""
        return self._username

    @property
    def claims(self) -> dict:
        """Returns the claims associated with these credentials."""
        return self._claims

    @property
    def roles(self) -> list[str]:
        """Returns the roles associated with these credentials."""
        return self._roles

    def to_header_value(self) -> str:
        """
        Returns the value to be used in the Authorization header for these credentials.

        Returns
        -------
        str
            The Authorization header value in the format
            "Basic base64(username:password)".
        """
        credentials = f"{self._username}:{self._password.get_value()}"
        encoded_credentials = base64.b64encode(credentials.encode("utf-8")).decode(
            "utf-8"
        )
        return f"Basic {encoded_credentials}"

    def match(self, username: str, password: str) -> bool:
        """
        Returns a value indicating whether the given username and password combination
        matches this credentials object. This method assumes that the username
        configured for this credentials object is a valid unicode string, and ignores
        encoding errors in the username received from the client.
        """
        return (
            secrets.compare_digest(
                self.username.encode("utf8"), username.encode("utf8", errors="ignore")
            )
            and self._password == password
        )


class BasicCredentialsProvider(ABC):
    """
    Abstract base class for providing Basic Authentication credentials dynamically.

    This class defines the interface for credential providers that can be used
    with BasicAuthentication to retrieve credentials at runtime. Implementations
    can fetch credentials from various sources such as databases, LDAP, configuration
    files, or external authentication services.
    """

    @abstractmethod
    async def get_credentials(self) -> list[BasicCredentials]:
        """
        Retrieve a list of valid Basic Authentication credentials.

        This method should return all currently valid credentials that can be used
        for authentication. The method is called by BasicAuthentication during
        the authentication process when no static credentials are provided.

        Returns
        -------
        list[BasicCredentials]
            A list of BasicCredentials instances representing all valid credentials
            for authentication. An empty list indicates no valid credentials are
            available.

        Notes
        -----
        - This method is called for each authentication attempt, so implementations
          should consider caching strategies for performance if fetching credentials is
          expensive.
        - The returned list should only contain currently valid and active credentials.
        - Consider implementing rate limiting and other security measures in your
          provider.
        """


class BasicAuthentication(AuthenticationHandler):
    """
    AuthenticationHandler that uses Basic Authentication for user authentication.

    This handler parses the Authorization header for Basic authentication credentials
    and validates them against configured credentials or a credentials provider.
    """

    def __init__(
        self,
        *credentials: BasicCredentials,
        scheme: str = "Basic",
        credentials_provider: BasicCredentialsProvider | None = None,
        description: str | None = None,
    ) -> None:
        """
        Creates a new instance of BasicAuthentication.

        Parameters
        ----------
        *credentials : BasicCredentials
            Static credentials handled by this instance.
        scheme: arbitrary scheme name of this authentication handler, default "Basic".
        credentials_provider : BasicCredentialsProvider | None, optional
            An optional provider that can be used to retrieve credentials dynamically.
            If not provided, only the static credentials will be used.
        description: optional description.
        Raises
        ------
        ValueError
            If neither credentials nor credentials_provider is provided.
        """
        super().__init__()
        self._scheme = scheme
        self._credentials: tuple[BasicCredentials, ...] = tuple(credentials)
        self._credentials_provider = credentials_provider
        self.description = description

        if not credentials and credentials_provider is None:
            raise ValueError(
                "Either credentials or credentials_provider must be provided"
            )

    @property
    def scheme(self) -> str:
        return self._scheme

    async def authenticate(self, context: Request) -> Identity | None:
        """
        Tries to authenticate the request using Basic Authentication.

        If authentication succeeds, returns an Identity with the authentication mode set
        to the scheme of the matching credentials. If no matching credentials are found,
        returns None.

        Parameters
        ----------
        context : Request
            The HTTP request context.

        Returns
        -------
        Identity | None
            An Identity object if authentication succeeds, None otherwise.
        """
        authorization_value = context.get_first_header(b"Authorization")

        if not authorization_value:
            return None

        if not authorization_value.startswith(b"Basic "):
            return None

        # Decode the base64 encoded credentials
        try:
            encoded_credentials = authorization_value[6:]  # Remove "Basic " prefix
            decoded_credentials = base64.b64decode(encoded_credentials).decode("utf-8")
            username, password = decoded_credentials.split(":", 1)
        except (ValueError, UnicodeDecodeError):
            # Invalid base64 or malformed credentials
            return None

        matching_credentials = await self._match_credentials(username, password)

        if matching_credentials is None:
            # The user provided a username and password combination, but they are
            # invalid. This kind of events must be logged and rate-limited to avoid the
            # risk of attackers trying to guess usernames and passwords.
            raise InvalidCredentialsError(context.original_client_ip)

        return self._get_identity_for_credentials(matching_credentials)

    def _get_identity_for_credentials(self, credentials: BasicCredentials) -> Identity:
        """
        Create an Identity object for the given credentials.

        Parameters
        ----------
        credentials : BasicCredentials
            The credentials to create an identity for.

        Returns
        -------
        Identity
            An Identity object containing the user's claims and roles.
        """
        claims = deepcopy(credentials.claims)
        claims.update(
            {"sub": credentials.username, "roles": [role for role in credentials.roles]}
        )
        return Identity(claims, authentication_mode=self.scheme)

    async def _match_credentials(
        self, username: str, password: str
    ) -> BasicCredentials | None:
        """
        Tries to find matching credentials for the given username and password.

        Parameters
        ----------
        username : str
            The provided username.
        password : str
            The provided password.

        Returns
        -------
        BasicCredentials | None
            The matching credentials if found, None otherwise.
        """
        credentials = self._credentials

        if not credentials and self._credentials_provider is not None:
            credentials = await self._credentials_provider.get_credentials()

        for cred in credentials:
            if cred.match(username, password):
                return cred

        return None



"""
This module contains high-level middlewares functions that can be used both by client
and server code.
"""

from enum import Enum
from typing import Awaitable, Callable, Iterable, overload

from blacksheep.messages import Response
from blacksheep.normalization import copy_special_attributes


def middleware_partial(handler, next_handler):
    async def middleware_wrapper(request):
        return await handler(request, next_handler)

    return middleware_wrapper


def get_middlewares_chain(middlewares, handler):
    fn = handler
    for middleware in reversed(middlewares):
        if not middleware:
            continue
        wrapper_fn = middleware_partial(middleware, fn)
        setattr(wrapper_fn, "root_fn", handler)
        copy_special_attributes(fn, wrapper_fn)
        fn = wrapper_fn
    return fn


class MiddlewareCategory(Enum):
    INIT = 10  # CORS, security headers, configuration that must happen early
    SESSION = 20  # Session handling
    AUTH = 30  # Authentication
    AUTHZ = 40  # Authorization
    BUSINESS = 50  # User business logic middlewares
    MESSAGE = 60  # Request/Response modification (default headers, etc.)


class CategorizedMiddleware:
    def __init__(
        self,
        middleware,
        category: MiddlewareCategory = MiddlewareCategory.BUSINESS,
        priority: int = 0,
    ):
        self.middleware = middleware
        self.category = category
        self.priority = priority


class MiddlewareList:
    """
    A list-like container for middlewares that supports categorized and
    prioritized insertion.
    """

    def __init__(self):
        self._middlewares: list[CategorizedMiddleware] = []
        self._is_sorted = True
        self._configured = False

    @overload
    def append(self, middleware: Callable[..., Awaitable[Response]]) -> None: ...

    @overload
    def append(
        self,
        middleware: Callable[..., Awaitable[Response]],
        category: MiddlewareCategory,
        priority: int = 0,
    ) -> None: ...

    def append(
        self,
        middleware: Callable[..., Awaitable[Response]],
        category: MiddlewareCategory = MiddlewareCategory.BUSINESS,
        priority: int = 0,
    ) -> None:
        """
        Add a middleware to the application.

        Args:
            middleware: The middleware function
            category: Where in the pipeline to place this middleware (default: BUSINESS)
            priority: Order within the category (lower = earlier, default: 0)
        """
        if self._configured:
            raise RuntimeError("Cannot add middlewares after configuration is complete")

        self._middlewares.append(CategorizedMiddleware(middleware, category, priority))
        self._is_sorted = False

    def insert(
        self, index: int, middleware: Callable[..., Awaitable[Response]]
    ) -> None:
        """Insert middleware at specific index (legacy support)"""
        if self._configured:
            raise RuntimeError("Cannot add middlewares after configuration is complete")

        # Convert to categorized middleware for consistency
        # This method defaults to MiddlewareCategory.INIT for backward compatibility,
        # as in the past insert was the only way to insert middlewares early in the
        # chain.
        self._middlewares.insert(
            index, CategorizedMiddleware(middleware, MiddlewareCategory.INIT, -1)
        )
        self._is_sorted = False

    def extend(self, middlewares) -> None:
        """Extend with multiple middlewares"""
        for middleware in middlewares:
            self.append(middleware)

    def clear(self) -> None:
        """Clear all middlewares"""
        self._middlewares.clear()
        self._is_sorted = True

    def _ensure_sorted(self) -> None:
        """Sort middlewares by category and priority if needed"""
        if not self._is_sorted:
            self._middlewares.sort(key=lambda m: (m.category.value, m.priority))
            self._is_sorted = True

    def _mark_configured(self) -> None:
        """Mark as configured to prevent further modifications"""
        self._configured = True
        self._ensure_sorted()

    def items(self) -> Iterable[CategorizedMiddleware]:
        yield from self._middlewares

    def __iter__(self):
        self._ensure_sorted()
        return iter(m.middleware for m in self._middlewares)

    def __len__(self) -> int:
        return len(self._middlewares)

    def __bool__(self) -> bool:
        return len(self._middlewares) > 0

    def __getitem__(self, index):
        self._ensure_sorted()
        return self._middlewares[index].middleware

    def to_list(self) -> list[Callable[..., Awaitable[Response]]]:
        """Get the sorted list of middleware functions"""
        self._ensure_sorted()
        return [m.middleware for m in self._middlewares]



"""
This module provides integration for OpenTelemetry using OTLP (OpenTelemetry Protocol)
exporters for logging and tracing in BlackSheep applications. This code is vendor
agnostic as it can work with all providers supporting the OpenTelemetry Protocol
(e.g. OpenTelemtry Collector, Grafana Alloy).

It defines a helper function to configure OpenTelemetry with OTLPLogExporter and
OTLPSpanExporter, ensuring that all required OTLP-related environment variables are set
before initialization.

Additional dependencies:
    pip install opentelemetry-exporter-otlp

Usage:
    from blacksheep.server.otel.otlp import use_open_telemetry_otlp

    use_open_telemetry_otlp(app)
"""

from opentelemetry.exporter.otlp.proto.http._log_exporter import OTLPLogExporter
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter

from blacksheep import Application

from . import OTELMiddleware, use_open_telemetry

__all__ = ["use_open_telemetry_otlp"]


def use_open_telemetry_otlp(app: Application, middleware: OTELMiddleware | None = None):
    """
    Configures OpenTelemetry for a BlackSheep application using OTLP exporters.

    This function configures OpenTelemetry logging and tracing using OTLPLogExporter
    and OTLPSpanExporter. It is your responsibility to configure OTEL env variables as
    desired:
    https://opentelemetry.io/docs/languages/sdk-configuration/otlp-exporter/

    Args:
        app: The BlackSheep Application instance.
        middleware (optional OTELMiddleware): Custom OTEL middleware instance.
            If not provided, the default OTELMiddleware is used.

    Raises:
        ValueError: If any required OTLP environment variables are missing.
    """
    use_open_telemetry(app, OTLPLogExporter(), OTLPSpanExporter(), middleware)



import pytest

from blacksheep import Request, Response
from blacksheep.contents import TextContent
from blacksheep.server.remotes.scheme import (
    HTTPSchemeMiddleware,
    configure_scheme_middleware,
)
from blacksheep.server.security.hsts import HSTSMiddleware
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend
from tests.utils.application import FakeApplication


class TestHTTPSchemeMiddleware:
    """Tests for HTTPSchemeMiddleware"""

    def test_middleware_accepts_http_scheme(self):
        middleware = HTTPSchemeMiddleware("http")
        assert middleware.scheme == "http"

    def test_middleware_accepts_https_scheme(self):
        middleware = HTTPSchemeMiddleware("https")
        assert middleware.scheme == "https"

    def test_middleware_rejects_invalid_scheme(self):
        with pytest.raises(TypeError, match="Invalid scheme, expected http | https"):
            HTTPSchemeMiddleware("ftp")

    async def test_middleware_forces_http_scheme(self):
        middleware = HTTPSchemeMiddleware("http")
        request = Request("GET", b"/", None)
        request.scheme = "https"  # Original scheme

        async def handler(req):
            assert req.scheme == "http"
            return Response(200)

        response = await middleware(request, handler)
        assert response.status == 200

    async def test_middleware_forces_https_scheme(self):
        middleware = HTTPSchemeMiddleware("https")
        request = Request("GET", b"/", None)
        request.scheme = "http"  # Original scheme

        async def handler(req):
            assert req.scheme == "https"
            return Response(200)

        response = await middleware(request, handler)
        assert response.status == 200


class TestConfigureSchemeMiddleware:
    """Tests for configure_scheme_middleware function"""

    @pytest.fixture
    def fake_app(self, monkeypatch):
        """Fixture that creates a FakeApplication with monkeypatch available"""

        def _create_app(**env_vars):
            for key, value in env_vars.items():
                monkeypatch.setenv(key, value)
            return FakeApplication()

        return _create_app

    async def test_force_https_enables_https_and_hsts(self, fake_app):
        app = fake_app(APP_FORCE_HTTPS="true")

        configure_scheme_middleware(app)

        # Check that HTTPSchemeMiddleware with https is added
        force_scheme_middlewares = [
            m for m in app.middlewares if isinstance(m, HTTPSchemeMiddleware)
        ]
        assert len(force_scheme_middlewares) == 1
        assert force_scheme_middlewares[0].scheme == "https"

        # Check that HSTSMiddleware is added
        hsts_middlewares = [m for m in app.middlewares if isinstance(m, HSTSMiddleware)]
        assert len(hsts_middlewares) == 1

    async def test_http_scheme_without_force_https(self, fake_app):
        app = fake_app(APP_HTTP_SCHEME="http")

        configure_scheme_middleware(app)

        # Check that HTTPSchemeMiddleware is added with http
        force_scheme_middlewares = [
            m for m in app.middlewares if isinstance(m, HTTPSchemeMiddleware)
        ]
        assert len(force_scheme_middlewares) == 1
        assert force_scheme_middlewares[0].scheme == "http"

        # Check that HSTSMiddleware is NOT added
        hsts_middlewares = [m for m in app.middlewares if isinstance(m, HSTSMiddleware)]
        assert len(hsts_middlewares) == 0

    async def test_https_scheme_without_force_https(self, fake_app):
        app = fake_app(APP_HTTP_SCHEME="https")

        configure_scheme_middleware(app)

        # Check that HTTPSchemeMiddleware is added with https
        force_scheme_middlewares = [
            m for m in app.middlewares if isinstance(m, HTTPSchemeMiddleware)
        ]
        assert len(force_scheme_middlewares) == 1
        assert force_scheme_middlewares[0].scheme == "https"

        # Check that HSTSMiddleware is NOT added (only when force_https=True)
        hsts_middlewares = [m for m in app.middlewares if isinstance(m, HSTSMiddleware)]
        assert len(hsts_middlewares) == 0

    async def test_no_middleware_when_not_configured(self, fake_app):
        app = fake_app()

        configure_scheme_middleware(app)

        # Check that no middleware is added
        force_scheme_middlewares = [
            m for m in app.middlewares if isinstance(m, HTTPSchemeMiddleware)
        ]
        assert len(force_scheme_middlewares) == 0

        hsts_middlewares = [m for m in app.middlewares if isinstance(m, HSTSMiddleware)]
        assert len(hsts_middlewares) == 0

    async def test_force_https_takes_precedence(self, fake_app):
        app = fake_app(APP_FORCE_HTTPS="true", APP_HTTP_SCHEME="http")

        configure_scheme_middleware(app)

        # Check that https is used despite http_scheme="http"
        force_scheme_middlewares = [
            m for m in app.middlewares if isinstance(m, HTTPSchemeMiddleware)
        ]
        assert len(force_scheme_middlewares) == 1
        assert force_scheme_middlewares[0].scheme == "https"

        # Check that HSTS is still added
        hsts_middlewares = [m for m in app.middlewares if isinstance(m, HSTSMiddleware)]
        assert len(hsts_middlewares) == 1


class TestEndToEndScheme:
    """End-to-end tests with real application requests"""

    async def test_request_with_forced_https_scheme(self, monkeypatch):
        monkeypatch.setenv("APP_FORCE_HTTPS", "true")
        app = FakeApplication()
        configure_scheme_middleware(app)

        @app.router.get("/")
        async def home(request: Request):
            return Response(200, content=TextContent(request.scheme))

        await app.start()

        scope = get_example_scope("GET", "/", [])
        scope["scheme"] = "http"  # Original scheme

        await app(scope, MockReceive(), MockSend())

        assert app.response.status == 200
        # The handler should see https due to middleware
        assert b"https" in app.response.content.body

    async def test_request_with_forced_http_scheme(self, monkeypatch):
        monkeypatch.setenv("APP_HTTP_SCHEME", "http")
        app = FakeApplication()
        configure_scheme_middleware(app)

        @app.router.get("/")
        async def home(request: Request):
            return Response(200, content=TextContent(request.scheme))

        await app.start()

        scope = get_example_scope("GET", "/", [])
        scope["scheme"] = "https"  # Original scheme

        await app(scope, MockReceive(), MockSend())

        assert app.response.status == 200
        # The handler should see http due to middleware
        assert b"http" in app.response.content.body



import sys
from dataclasses import dataclass
from datetime import date, datetime
from enum import IntEnum
from typing import Generic, Mapping, Sequence, TypeVar, Union
from uuid import UUID

if sys.version_info >= (3, 9):
    from typing import Annotated

import pytest
from openapidocs.common import Format, Serializer
from openapidocs.v3 import (
    APIKeySecurity,
    HTTPSecurity,
    Info,
    OAuth2Security,
    OAuthFlow,
    OAuthFlows,
    OpenIdConnectSecurity,
    ParameterLocation,
    Reference,
    Schema,
    ValueType,
)
from pydantic import VERSION as PYDANTIC_LIB_VERSION
from pydantic import BaseModel, HttpUrl
from pydantic.types import (
    UUID4,
    NegativeFloat,
    PositiveInt,
    condecimal,
    confloat,
    conint,
)

from blacksheep.messages import Response
from blacksheep.server.application import Application
from blacksheep.server.bindings import FromForm
from blacksheep.server.controllers import APIController
from blacksheep.server.openapi.common import (
    ContentInfo,
    DefaultSerializer,
    EndpointDocs,
    OpenAPIEndpointException,
    ResponseInfo,
    SecurityInfo,
)
from blacksheep.server.openapi.exceptions import DuplicatedContentTypeDocsException
from blacksheep.server.openapi.v3 import (
    DataClassTypeHandler,
    OpenAPIHandler,
    PydanticModelTypeHandler,
    Tag,
    check_union,
)
from blacksheep.server.routing import RoutesRegistry

GenericModel = BaseModel

PYDANTIC_VERSION = 2

if int(PYDANTIC_LIB_VERSION[0]) < 2:
    from pydantic.generics import GenericModel

    PYDANTIC_VERSION = 1

try:
    from pydantic import field_validator
except ImportError:
    # Pydantic v1
    from pydantic import validator as field_validator


T = TypeVar("T")
U = TypeVar("U")


@dataclass
class DateTest:
    one: date
    two: datetime


class PydExampleWithSpecificTypes(BaseModel):
    url: HttpUrl


class PydCat(BaseModel):
    id: int
    name: str
    childs: list[UUID4]


class PydPaginatedSetOfCat(BaseModel):
    items: list[PydCat]
    total: int


class PydTypeWithChildModels(BaseModel):
    child: PydPaginatedSetOfCat
    friend: PydExampleWithSpecificTypes


class PlainClass:
    id: int
    name: str


@dataclass
class Cat:
    id: int
    name: str


@dataclass
class CreateCatInput:
    name: str


@dataclass
class CatOwner:
    id: int
    first_name: str
    last_name: str


@dataclass
class CatDetails(Cat):
    owner: CatOwner
    friends: list[int]


@dataclass
class CreateCatImages:
    images: list[str]


@dataclass
class Combo(Generic[T, U]):
    item_one: T
    item_two: U


@dataclass
class PaginatedSet(Generic[T]):
    items: list[T]
    total: int


@dataclass
class Validated(Generic[T]):
    data: T
    error: str


@dataclass
class SubValidated(Generic[T]):
    sub: Validated[T]


class FooLevel(IntEnum):
    BASIC = 1
    MEDIUM = 2
    SUPER = 3


@dataclass
class Foo:
    a: str
    b: bool
    level: FooLevel | None = None


@dataclass
class CreateFooInput:
    a: str
    b: bool
    level: FooLevel | None = None


@dataclass
class Ufo:
    b: bool
    c: str


@dataclass
class ForwardRefExample:
    value: "PaginatedSet[Cat]"


@dataclass
class GenericWithForwardRefExample(Generic[T]):
    ufo: "Ufo"
    value: "PaginatedSet[T]"


class Error(BaseModel):
    code: int
    message: str


class DataModel(BaseModel):
    numbers: list[int]
    people: list[str]


class PydResponse(GenericModel, Generic[T]):
    data: T | None
    error: Error | None

    @field_validator("error")
    def check_consistency(cls, v, values):
        if v is not None and values["data"] is not None:
            raise ValueError("must not provide both data and error")
        if v is None and values.get("data") is None:
            raise ValueError("must provide data or error")
        return v


class PydConstrained(BaseModel):
    a: PositiveInt
    b: NegativeFloat
    big_int: conint(gt=1000, lt=1024)

    big_float: confloat(gt=1000.0, lt=1024.0)
    unit_interval: confloat(ge=0, le=1)

    decimal_positive: condecimal(gt=0)
    decimal_negative: condecimal(lt=0)


def get_app() -> Application:
    app = Application()
    app.controllers_router = RoutesRegistry()
    return app


def get_cats_api() -> Application:
    app = get_app()
    get = app.router.get
    post = app.router.post
    delete = app.router.delete

    @get("/api/cats")
    def get_cats() -> PaginatedSet[Cat]: ...

    @get("/api/cats/{cat_id}")
    def get_cat_details(cat_id: int) -> CatDetails: ...

    @post("/api/cats")
    def create_cat(input: CreateCatInput) -> Cat: ...

    @delete("/api/cats/{cat_id}")
    def delete_cat(cat_id: int) -> None: ...

    @post("/api/cats/{cat_id}/images")
    def upload_images(cat_id: int, images: FromForm[CreateCatImages]) -> None: ...

    return app


def get_cats_annotated_api() -> Application:
    app = get_app()
    get = app.router.get
    post = app.router.post
    delete = app.router.delete

    @get("/api/cats")
    def get_cats_annotated() -> Annotated[Response, PaginatedSet[Cat]]: ...

    @get("/api/cats_2")
    def get_cats_annotated_2() -> Annotated[Response, PaginatedSet[Cat], list[Cat]]: ...

    @get("/api/cats_3")
    def get_cats_annotated_3() -> (
        Annotated[Response, PaginatedSet[Cat] | list[Cat]]
    ): ...

    @get("/api/cats/{cat_id}")
    def get_cat_details_annotated(cat_id: int) -> Annotated[Response, CatDetails]: ...

    @post("/api/cats")
    def create_cat_annotated(input: CreateCatInput) -> Annotated[Response, Cat]: ...

    @delete("/api/cats/{cat_id}")
    def delete_cat_annotated(cat_id: int) -> Annotated[Response, None]: ...

    return app


@pytest.fixture(scope="function")
def docs() -> OpenAPIHandler:
    # example documentation
    return OpenAPIHandler(info=Info("Example", "0.0.1"))


class CapitalizeOperationDocs(OpenAPIHandler):
    def get_operation_id(self, docs: EndpointDocs | None, handler) -> str:
        return handler.__name__.capitalize().replace("_", " ")


@pytest.fixture
def capitalize_operation_id_docs() -> CapitalizeOperationDocs:
    return CapitalizeOperationDocs(info=Info("Example", "0.0.1"))


@pytest.fixture
def serializer() -> Serializer:
    return Serializer()


async def test_raises_for_started_app(docs):
    app = get_app()

    await app.start()

    with pytest.raises(TypeError):
        docs.bind_app(app)


async def test_raises_for_duplicated_content_example(docs):
    app = get_app()

    @app.router.get("/")
    @docs(
        responses={
            200: ResponseInfo("Example", content=[ContentInfo(Foo), ContentInfo(Foo)])
        }
    )
    async def example(): ...

    with pytest.raises(DuplicatedContentTypeDocsException):
        docs.bind_app(app)
        await app.start()


@pytest.mark.parametrize(
    "annotation,expected_result",
    [
        (Foo, [False, Foo]),
        (Foo | None, [True, Foo]),
        (Foo | None, [True, Foo]),
        (None | Foo, [True, Foo]),
    ],
)
def test_check_union(annotation, expected_result):
    assert check_union(annotation) == tuple(expected_result)


def test_register_schema_can_handle_classes_with_same_name(docs):
    @dataclass
    class FooX:
        x: str

    FooX.__name__ = "Foo"

    docs.register_schema_for_type(Foo)
    docs.register_schema_for_type(FooX)

    foo_schema = docs.components.schemas["Foo"]
    foox_schema = docs.components.schemas["Foo1"]

    assert foo_schema is not None
    assert foox_schema is not None

    assert isinstance(foo_schema, Schema)
    assert isinstance(foox_schema, Schema)

    assert "x" in foox_schema.properties
    assert "a" in foo_schema.properties


def test_register_schema_handles_repeated_calls(docs):
    docs.register_schema_for_type(Foo)
    docs.register_schema_for_type(Foo)
    docs.register_schema_for_type(Foo)

    assert docs.components.schemas is not None
    assert len(docs.components.schemas) == 1
    foo_schema = docs.components.schemas["Foo"]
    assert "Foo1" not in docs.components.schemas

    assert foo_schema is not None
    assert isinstance(foo_schema, Schema)
    assert "a" in foo_schema.properties


def test_handles_forward_refs(docs):
    @dataclass
    class Friend:
        foo: "Foo"

    docs.register_schema_for_type(Foo)
    docs.register_schema_for_type(Friend)

    assert docs.components.schemas is not None
    assert len(docs.components.schemas) == 2
    friend_schema = docs.components.schemas["Friend"]

    assert friend_schema is not None
    assert isinstance(friend_schema, Schema)
    assert friend_schema.properties["foo"] == Reference(ref="#/components/schemas/Foo")


def test_register_schema_for_enum(docs):
    docs.register_schema_for_type(FooLevel)

    assert docs.components.schemas is not None
    assert len(docs.components.schemas) == 1
    schema = docs.components.schemas["FooLevel"]

    assert isinstance(schema, Schema)

    assert schema is not None
    assert schema.type == ValueType.INTEGER
    assert schema.enum == [x.value for x in FooLevel]


def test_try_get_schema_for_enum_returns_none_for_not_enum(docs):
    assert docs._try_get_schema_for_enum(Foo) is None


def test_get_parameters_returns_non_for_object_without_binders(docs):
    assert docs.get_parameters(Foo) is None
    assert docs.get_request_body(Foo) is None


def test_get_content_from_response_info_returns_none_for_missing_content(docs):
    assert docs._get_content_from_response_info(None) is None


def test_open_api_handler_object_handlers(docs: OpenAPIHandler):
    assert docs.object_types_handlers is not None
    assert isinstance(docs.object_types_handlers[0], DataClassTypeHandler)
    assert isinstance(docs.object_types_handlers[1], PydanticModelTypeHandler)


def test_open_api_handler_get_fields(docs: OpenAPIHandler):
    assert docs.get_fields(None) == []


@pytest.mark.parametrize(
    "json_path,yaml_path,preferred_format,expected_result",
    [
        ["foo.json", "foo.yaml", Format.YAML, "foo.yaml"],
        ["foo.json", "foo.yaml", Format.JSON, "foo.json"],
        ["/openapi.json", "/openapi.yaml", Format.YAML, "/openapi.yaml"],
    ],
)
def test_get_spec_path_preferred_format(
    json_path, yaml_path, preferred_format, expected_result
):
    docs = OpenAPIHandler(
        info=Info("Example", "0.0.1"),
        json_spec_path=json_path,
        yaml_spec_path=yaml_path,
        preferred_format=preferred_format,
    )

    assert docs.get_spec_path() == expected_result


def test_get_spec_path_raises_for_unsupported_preferred_format(docs):
    docs.preferred_format = "NOPE"  # type: ignore

    with pytest.raises(OpenAPIEndpointException):
        docs.get_spec_path()


def test_dates_handling(docs: OpenAPIHandler, serializer: Serializer):
    docs.register_schema_for_type(DateTest)

    assert docs.components.schemas is not None
    schema = docs.components.schemas["DateTest"]

    assert isinstance(schema, Schema)

    yaml = serializer.to_yaml(docs.generate_documentation(get_app()))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths: {}
components:
    schemas:
        DateTest:
            type: object
            required:
            - one
            - two
            properties:
                one:
                    type: string
                    format: date
                    nullable: false
                two:
                    type: string
                    format: date-time
                    nullable: false
tags: []
""".strip()
    )


def test_register_schema_for_generic_with_list(
    docs: OpenAPIHandler, serializer: Serializer
):
    docs.register_schema_for_type(PaginatedSet[Foo])

    assert docs.components.schemas is not None
    schema = docs.components.schemas["PaginatedSetOfFoo"]

    assert isinstance(schema, Schema)

    yaml = serializer.to_yaml(docs.generate_documentation(get_app()))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths: {}
components:
    schemas:
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        PaginatedSetOfFoo:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Foo'
                total:
                    type: integer
                    format: int64
                    nullable: false
tags: []
""".strip()
    )


def test_register_schema_for_multiple_generic_with_list(
    docs: OpenAPIHandler, serializer: Serializer
):
    docs.register_schema_for_type(PaginatedSet[Foo])
    docs.register_schema_for_type(PaginatedSet[Ufo])

    assert docs.components.schemas is not None
    schema = docs.components.schemas["PaginatedSetOfFoo"]
    assert isinstance(schema, Schema)

    schema = docs.components.schemas["PaginatedSetOfUfo"]
    assert isinstance(schema, Schema)

    yaml = serializer.to_yaml(docs.generate_documentation(get_app()))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths: {}
components:
    schemas:
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        PaginatedSetOfFoo:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Foo'
                total:
                    type: integer
                    format: int64
                    nullable: false
        Ufo:
            type: object
            required:
            - b
            - c
            properties:
                b:
                    type: boolean
                    nullable: false
                c:
                    type: string
                    nullable: false
        PaginatedSetOfUfo:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Ufo'
                total:
                    type: integer
                    format: int64
                    nullable: false
tags: []
""".strip()
    )


def test_register_schema_for_generic_with_property(
    docs: OpenAPIHandler, serializer: Serializer
):
    docs.register_schema_for_type(Validated[Foo])

    assert docs.components.schemas is not None
    schema = docs.components.schemas["ValidatedOfFoo"]

    assert isinstance(schema, Schema)

    yaml = serializer.to_yaml(docs.generate_documentation(get_app()))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths: {}
components:
    schemas:
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        ValidatedOfFoo:
            type: object
            required:
            - data
            - error
            properties:
                data:
                    $ref: '#/components/schemas/Foo'
                error:
                    type: string
                    nullable: false
tags: []
""".strip()
    )


def test_register_schema_for_generic_sub_property(
    docs: OpenAPIHandler, serializer: Serializer
):
    docs.register_schema_for_type(Validated[Foo])
    docs.register_schema_for_type(SubValidated[Foo])

    assert docs.components.schemas is not None
    schema = docs.components.schemas["SubValidatedOfFoo"]

    assert isinstance(schema, Schema)

    yaml = serializer.to_yaml(docs.generate_documentation(get_app()))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths: {}
components:
    schemas:
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        ValidatedOfFoo:
            type: object
            required:
            - data
            - error
            properties:
                data:
                    $ref: '#/components/schemas/Foo'
                error:
                    type: string
                    nullable: false
        SubValidatedOfFoo:
            type: object
            required:
            - sub
            properties:
                sub:
                    $ref: '#/components/schemas/ValidatedOfFoo'
tags: []
""".strip()
    )


async def test_register_schema_for_multi_generic(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.router.route("/combo")
    def combo_example() -> Combo[Cat, Foo]: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /combo:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/ComboOfCatAndFoo'
            operationId: combo_example
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        ComboOfCatAndFoo:
            type: object
            required:
            - item_one
            - item_two
            properties:
                item_one:
                    $ref: '#/components/schemas/Cat'
                item_two:
                    $ref: '#/components/schemas/Foo'
tags: []
""".strip()
    )


async def test_register_schema_for_generic_with_list_reusing_ref(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()
    docs.bind_app(app)

    @docs(tags=["B tag"])
    @app.router.route("/one")
    def one() -> PaginatedSet[Cat]: ...

    @docs(tags=["A tag"])
    @app.router.route("/two")
    def two() -> PaginatedSet[Cat]: ...

    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /one:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfCat'
            tags:
            - B tag
            operationId: one
    /two:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfCat'
            tags:
            - A tag
            operationId: two
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Cat'
                total:
                    type: integer
                    format: int64
                    nullable: false
tags:
-   name: A tag
-   name: B tag
""".strip()
    )


def test_get_type_name_raises_for_invalid_object_type(docs: OpenAPIHandler):
    with pytest.raises(ValueError):
        docs.get_type_name(10)


async def test_handling_of_forward_references(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.router.route("/")
    def forward_ref_example() -> ForwardRefExample: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/ForwardRefExample'
            operationId: forward_ref_example
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Cat'
                total:
                    type: integer
                    format: int64
                    nullable: false
        ForwardRefExample:
            type: object
            required:
            - value
            properties:
                value:
                    $ref: '#/components/schemas/PaginatedSetOfCat'
tags: []
""".strip()
    )


async def test_handling_of_normal_class(docs: OpenAPIHandler, serializer: Serializer):
    """
    Plain classes are simply ignored, since their handling would be ambiguous:
    should the library inspect type annotations, __dict__?
    (in fact, the built-in json module throws for them).
    """
    app = get_app()

    @app.router.route("/")
    def plain_class() -> PlainClass: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                nullable: false
            operationId: plain_class
components: {}
tags: []
""".strip()
    )


async def test_handling_of_pydantic_class_with_generic(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.router.route("/")
    def home() -> PydPaginatedSetOfCat: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    if PYDANTIC_VERSION == 1:
        assert (
            yaml.strip()
            == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydPaginatedSetOfCat'
            operationId: home
components:
    schemas:
        PydCat:
            title: PydCat
            type: object
            properties:
                id:
                    title: Id
                    type: integer
                name:
                    title: Name
                    type: string
                childs:
                    title: Childs
                    type: array
                    items:
                        type: string
                        format: uuid4
            required:
            - id
            - name
            - childs
        PydPaginatedSetOfCat:
            title: PydPaginatedSetOfCat
            type: object
            properties:
                items:
                    title: Items
                    type: array
                    items:
                        $ref: '#/components/schemas/PydCat'
                total:
                    title: Total
                    type: integer
            required:
            - items
            - total
tags: []
""".strip()
        )
    else:
        assert (
            yaml.strip()
            == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydPaginatedSetOfCat'
            operationId: home
components:
    schemas:
        PydCat:
            properties:
                id:
                    title: Id
                    type: integer
                name:
                    title: Name
                    type: string
                childs:
                    items:
                        format: uuid4
                        type: string
                    title: Childs
                    type: array
            required:
            - id
            - name
            - childs
            title: PydCat
            type: object
        PydPaginatedSetOfCat:
            properties:
                items:
                    items:
                        $ref: '#/components/schemas/PydCat'
                    title: Items
                    type: array
                total:
                    title: Total
                    type: integer
            required:
            - items
            - total
            title: PydPaginatedSetOfCat
            type: object
tags: []
""".strip()
        )


async def test_handling_of_pydantic_class_with_child_models(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.router.route("/")
    def home() -> PydTypeWithChildModels: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    if PYDANTIC_VERSION == 1:
        assert (
            yaml.strip()
            == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydTypeWithChildModels'
            operationId: home
components:
    schemas:
        PydCat:
            title: PydCat
            type: object
            properties:
                id:
                    title: Id
                    type: integer
                name:
                    title: Name
                    type: string
                childs:
                    title: Childs
                    type: array
                    items:
                        type: string
                        format: uuid4
            required:
            - id
            - name
            - childs
        PydPaginatedSetOfCat:
            title: PydPaginatedSetOfCat
            type: object
            properties:
                items:
                    title: Items
                    type: array
                    items:
                        $ref: '#/components/schemas/PydCat'
                total:
                    title: Total
                    type: integer
            required:
            - items
            - total
        PydExampleWithSpecificTypes:
            title: PydExampleWithSpecificTypes
            type: object
            properties:
                url:
                    title: Url
                    minLength: 1
                    maxLength: 2083
                    format: uri
                    type: string
            required:
            - url
        PydTypeWithChildModels:
            title: PydTypeWithChildModels
            type: object
            properties:
                child:
                    $ref: '#/components/schemas/PydPaginatedSetOfCat'
                friend:
                    $ref: '#/components/schemas/PydExampleWithSpecificTypes'
            required:
            - child
            - friend
tags: []
    """.strip()
        )
    else:
        assert (
            yaml.strip()
            == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydTypeWithChildModels'
            operationId: home
components:
    schemas:
        PydCat:
            properties:
                id:
                    title: Id
                    type: integer
                name:
                    title: Name
                    type: string
                childs:
                    items:
                        format: uuid4
                        type: string
                    title: Childs
                    type: array
            required:
            - id
            - name
            - childs
            title: PydCat
            type: object
        PydExampleWithSpecificTypes:
            properties:
                url:
                    format: uri
                    maxLength: 2083
                    minLength: 1
                    title: Url
                    type: string
            required:
            - url
            title: PydExampleWithSpecificTypes
            type: object
        PydPaginatedSetOfCat:
            properties:
                items:
                    items:
                        $ref: '#/components/schemas/PydCat'
                    title: Items
                    type: array
                total:
                    title: Total
                    type: integer
            required:
            - items
            - total
            title: PydPaginatedSetOfCat
            type: object
        PydTypeWithChildModels:
            properties:
                child:
                    $ref: '#/components/schemas/PydPaginatedSetOfCat'
                friend:
                    $ref: '#/components/schemas/PydExampleWithSpecificTypes'
            required:
            - child
            - friend
            title: PydTypeWithChildModels
            type: object
tags: []
    """.strip()
        )


async def test_handling_of_pydantic_class_in_generic(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.router.route("/")
    def home() -> PaginatedSet[PydCat]: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    if PYDANTIC_VERSION == 1:
        assert (
            yaml.strip()
            == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfPydCat'
            operationId: home
components:
    schemas:
        PydCat:
            title: PydCat
            type: object
            properties:
                id:
                    title: Id
                    type: integer
                name:
                    title: Name
                    type: string
                childs:
                    title: Childs
                    type: array
                    items:
                        type: string
                        format: uuid4
            required:
            - id
            - name
            - childs
        PaginatedSetOfPydCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/PydCat'
                total:
                    type: integer
                    format: int64
                    nullable: false
tags: []
    """.strip()
        )
    else:
        assert (
            yaml.strip()
            == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfPydCat'
            operationId: home
components:
    schemas:
        PydCat:
            properties:
                id:
                    title: Id
                    type: integer
                name:
                    title: Name
                    type: string
                childs:
                    items:
                        format: uuid4
                        type: string
                    title: Childs
                    type: array
            required:
            - id
            - name
            - childs
            title: PydCat
            type: object
        PaginatedSetOfPydCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/PydCat'
                total:
                    type: integer
                    format: int64
                    nullable: false
tags: []
""".strip()
        )


async def test_handling_of_sequence(docs: OpenAPIHandler, serializer: Serializer):
    app = get_app()

    @app.router.route("/")
    def home() -> Sequence[Cat]: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                type: array
                                nullable: false
                                items:
                                    $ref: '#/components/schemas/Cat'
            operationId: home
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
tags: []
""".strip()
    )


@pytest.mark.asyncio
async def test_handling_of_mapping(docs: OpenAPIHandler, serializer: Serializer):
    app = get_app()

    @app.router.route("/")
    def home() -> Mapping[str, Mapping[int, list[Cat]]]: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == r"""
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                type: object
                                additionalProperties:
                                    type: object
                                    additionalProperties:
                                        type: array
                                        nullable: false
                                        items:
                                            $ref: '#/components/schemas/Cat'
                                    nullable: false
                                nullable: false
            operationId: home
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
tags: []
""".strip()
    )


def test_handling_of_generic_with_forward_references(docs: OpenAPIHandler):
    with pytest.warns(UserWarning):
        docs.register_schema_for_type(GenericWithForwardRefExample[Cat])


async def test_cats_api(docs: OpenAPIHandler, serializer: Serializer):
    app = get_cats_api()
    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /api/cats:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfCat'
            operationId: get_cats
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Cat'
            operationId: create_cat
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/CreateCatInput'
                required: true
    /api/cats/{cat_id}:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/CatDetails'
            operationId: get_cat_details
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
        delete:
            responses:
                '204':
                    description: Success response
            operationId: delete_cat
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
    /api/cats/{cat_id}/images:
        post:
            responses:
                '204':
                    description: Success response
            operationId: upload_images
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
            requestBody:
                content:
                    multipart/form-data:
                        schema:
                            $ref: '#/components/schemas/CreateCatImages'
                    application/x-www-form-urlencoded:
                        schema:
                            $ref: '#/components/schemas/CreateCatImages'
                required: true
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Cat'
                total:
                    type: integer
                    format: int64
                    nullable: false
        CreateCatInput:
            type: object
            required:
            - name
            properties:
                name:
                    type: string
                    nullable: false
        CatOwner:
            type: object
            required:
            - id
            - first_name
            - last_name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                first_name:
                    type: string
                    nullable: false
                last_name:
                    type: string
                    nullable: false
        CatDetails:
            type: object
            required:
            - id
            - name
            - owner
            - friends
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
                owner:
                    $ref: '#/components/schemas/CatOwner'
                friends:
                    type: array
                    nullable: false
                    items:
                        type: integer
                        format: int64
                        nullable: false
        CreateCatImages:
            type: object
            required:
            - images
            properties:
                images:
                    type: array
                    nullable: false
                    items:
                        type: string
                        nullable: false
tags: []
""".strip()
    )


async def test_cats_api_capital_operations_ids(
    capitalize_operation_id_docs: CapitalizeOperationDocs,
    serializer: Serializer,
):
    app = get_cats_api()
    docs = capitalize_operation_id_docs

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /api/cats:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfCat'
            operationId: Get cats
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Cat'
            operationId: Create cat
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/CreateCatInput'
                required: true
    /api/cats/{cat_id}:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/CatDetails'
            operationId: Get cat details
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
        delete:
            responses:
                '204':
                    description: Success response
            operationId: Delete cat
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
    /api/cats/{cat_id}/images:
        post:
            responses:
                '204':
                    description: Success response
            operationId: Upload images
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
            requestBody:
                content:
                    multipart/form-data:
                        schema:
                            $ref: '#/components/schemas/CreateCatImages'
                    application/x-www-form-urlencoded:
                        schema:
                            $ref: '#/components/schemas/CreateCatImages'
                required: true
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Cat'
                total:
                    type: integer
                    format: int64
                    nullable: false
        CreateCatInput:
            type: object
            required:
            - name
            properties:
                name:
                    type: string
                    nullable: false
        CatOwner:
            type: object
            required:
            - id
            - first_name
            - last_name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                first_name:
                    type: string
                    nullable: false
                last_name:
                    type: string
                    nullable: false
        CatDetails:
            type: object
            required:
            - id
            - name
            - owner
            - friends
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
                owner:
                    $ref: '#/components/schemas/CatOwner'
                friends:
                    type: array
                    nullable: false
                    items:
                        type: integer
                        format: int64
                        nullable: false
        CreateCatImages:
            type: object
            required:
            - images
            properties:
                images:
                    type: array
                    nullable: false
                    items:
                        type: string
                        nullable: false
tags: []
""".strip()
    )


async def test_cats_annotated_api(docs: OpenAPIHandler, serializer: Serializer):
    app = get_cats_annotated_api()
    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /api/cats:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfCat'
            operationId: get_cats_annotated
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Cat'
            operationId: create_cat_annotated
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/CreateCatInput'
                required: true
    /api/cats_2:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/ResponseOfPaginatedSetOfCatAndListOfCat'
            operationId: get_cats_annotated_2
    /api/cats_3:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/UnionOfPaginatedSetOfCatAndListOfCat'
            operationId: get_cats_annotated_3
    /api/cats/{cat_id}:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/CatDetails'
            operationId: get_cat_details_annotated
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
        delete:
            responses:
                '204':
                    description: Success response
            operationId: delete_cat_annotated
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Cat'
                total:
                    type: integer
                    format: int64
                    nullable: false
        CreateCatInput:
            type: object
            required:
            - name
            properties:
                name:
                    type: string
                    nullable: false
        ResponseOfPaginatedSetOfCatAndListOfCat:
            type: object
            anyOf:
            -   $ref: '#/components/schemas/PaginatedSetOfCat'
            -   type: array
                nullable: false
                items:
                    $ref: '#/components/schemas/Cat'
        UnionOfPaginatedSetOfCatAndListOfCat:
            type: object
            anyOf:
            -   $ref: '#/components/schemas/PaginatedSetOfCat'
            -   type: array
                nullable: false
                items:
                    $ref: '#/components/schemas/Cat'
        CatOwner:
            type: object
            required:
            - id
            - first_name
            - last_name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                first_name:
                    type: string
                    nullable: false
                last_name:
                    type: string
                    nullable: false
        CatDetails:
            type: object
            required:
            - id
            - name
            - owner
            - friends
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
                owner:
                    $ref: '#/components/schemas/CatOwner'
                friends:
                    type: array
                    nullable: false
                    items:
                        type: integer
                        format: int64
                        nullable: false
tags: []
""".strip()
    )


async def test_cats_annotated_api_capital_operations_ids(
    capitalize_operation_id_docs: CapitalizeOperationDocs,
    serializer: Serializer,
):
    app = get_cats_annotated_api()
    docs = capitalize_operation_id_docs

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /api/cats:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfCat'
            operationId: Get cats annotated
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Cat'
            operationId: Create cat annotated
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/CreateCatInput'
                required: true
    /api/cats_2:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/ResponseOfPaginatedSetOfCatAndListOfCat'
            operationId: Get cats annotated 2
    /api/cats_3:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/UnionOfPaginatedSetOfCatAndListOfCat'
            operationId: Get cats annotated 3
    /api/cats/{cat_id}:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/CatDetails'
            operationId: Get cat details annotated
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
        delete:
            responses:
                '204':
                    description: Success response
            operationId: Delete cat annotated
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Cat'
                total:
                    type: integer
                    format: int64
                    nullable: false
        CreateCatInput:
            type: object
            required:
            - name
            properties:
                name:
                    type: string
                    nullable: false
        ResponseOfPaginatedSetOfCatAndListOfCat:
            type: object
            anyOf:
            -   $ref: '#/components/schemas/PaginatedSetOfCat'
            -   type: array
                nullable: false
                items:
                    $ref: '#/components/schemas/Cat'
        UnionOfPaginatedSetOfCatAndListOfCat:
            type: object
            anyOf:
            -   $ref: '#/components/schemas/PaginatedSetOfCat'
            -   type: array
                nullable: false
                items:
                    $ref: '#/components/schemas/Cat'
        CatOwner:
            type: object
            required:
            - id
            - first_name
            - last_name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                first_name:
                    type: string
                    nullable: false
                last_name:
                    type: string
                    nullable: false
        CatDetails:
            type: object
            required:
            - id
            - name
            - owner
            - friends
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
                owner:
                    $ref: '#/components/schemas/CatOwner'
                friends:
                    type: array
                    nullable: false
                    items:
                        type: integer
                        format: int64
                        nullable: false
tags: []
""".strip()
    )


async def test_handling_of_pydantic_types(docs: OpenAPIHandler, serializer: Serializer):
    app = get_app()

    @app.router.route("/")
    def home() -> PydExampleWithSpecificTypes: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    if PYDANTIC_VERSION == 1:
        assert (
            yaml.strip()
            == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydExampleWithSpecificTypes'
            operationId: home
components:
    schemas:
        PydExampleWithSpecificTypes:
            title: PydExampleWithSpecificTypes
            type: object
            properties:
                url:
                    title: Url
                    minLength: 1
                    maxLength: 2083
                    format: uri
                    type: string
            required:
            - url
tags: []
""".strip()
        )
        return

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydExampleWithSpecificTypes'
            operationId: home
components:
    schemas:
        PydExampleWithSpecificTypes:
            properties:
                url:
                    format: uri
                    maxLength: 2083
                    minLength: 1
                    title: Url
                    type: string
            required:
            - url
            title: PydExampleWithSpecificTypes
            type: object
tags: []
""".strip()
    )


async def test_pydantic_generic(docs: OpenAPIHandler, serializer: Serializer):
    app = get_app()

    @app.router.route("/")
    def home() -> PydResponse[PydCat]: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    if PYDANTIC_VERSION == 1:
        expected_result = """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydResponse[PydCat]'
            operationId: home
components:
    schemas:
        PydCat:
            title: PydCat
            type: object
            properties:
                id:
                    title: Id
                    type: integer
                name:
                    title: Name
                    type: string
                childs:
                    title: Childs
                    type: array
                    items:
                        type: string
                        format: uuid4
            required:
            - id
            - name
            - childs
        Error:
            title: Error
            type: object
            properties:
                code:
                    title: Code
                    type: integer
                message:
                    title: Message
                    type: string
            required:
            - code
            - message
        PydResponse[PydCat]:
            title: PydResponse[PydCat]
            type: object
            properties:
                data:
                    $ref: '#/components/schemas/PydCat'
                error:
                    $ref: '#/components/schemas/Error'
tags: []
""".strip()
    elif PYDANTIC_VERSION == 2:
        expected_result = """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydResponse[PydCat]'
            operationId: home
components:
    schemas:
        Error:
            properties:
                code:
                    title: Code
                    type: integer
                message:
                    title: Message
                    type: string
            required:
            - code
            - message
            title: Error
            type: object
        PydCat:
            properties:
                id:
                    title: Id
                    type: integer
                name:
                    title: Name
                    type: string
                childs:
                    items:
                        format: uuid4
                        type: string
                    title: Childs
                    type: array
            required:
            - id
            - name
            - childs
            title: PydCat
            type: object
        PydResponse[PydCat]:
            properties:
                data:
                    anyOf:
                    -   $ref: '#/components/schemas/PydCat'
                    -   type: 'null'
                error:
                    anyOf:
                    -   $ref: '#/components/schemas/Error'
                    -   type: 'null'
            required:
            - data
            - error
            title: PydResponse[PydCat]
            type: object
tags: []
""".strip()
    else:
        raise RuntimeError("Missing expected_result")
    assert yaml.strip() == expected_result


async def test_pydantic_constrained_types(docs: OpenAPIHandler, serializer: Serializer):
    app = get_app()

    @app.router.route("/")
    def home() -> PydConstrained: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))
    expected_result: str

    if PYDANTIC_VERSION == 1:
        expected_result = """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydConstrained'
            operationId: home
components:
    schemas:
        PydConstrained:
            title: PydConstrained
            type: object
            properties:
                a:
                    title: A
                    exclusiveMinimum: 0
                    type: integer
                b:
                    title: B
                    exclusiveMaximum: 0
                    type: number
                big_int:
                    title: Big Int
                    exclusiveMinimum: 1000
                    exclusiveMaximum: 1024
                    type: integer
                big_float:
                    title: Big Float
                    exclusiveMinimum: 1000.0
                    exclusiveMaximum: 1024.0
                    type: number
                unit_interval:
                    title: Unit Interval
                    minimum: 0
                    maximum: 1
                    type: number
                decimal_positive:
                    title: Decimal Positive
                    exclusiveMinimum: 0
                    type: number
                decimal_negative:
                    title: Decimal Negative
                    exclusiveMaximum: 0
                    type: number
            required:
            - a
            - b
            - big_int
            - big_float
            - unit_interval
            - decimal_positive
            - decimal_negative
tags: []
""".strip()
    elif PYDANTIC_VERSION == 2:
        expected_result = r"""
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydConstrained'
            operationId: home
components:
    schemas:
        PydConstrained:
            properties:
                a:
                    exclusiveMinimum: 0
                    title: A
                    type: integer
                b:
                    exclusiveMaximum: 0
                    title: B
                    type: number
                big_int:
                    exclusiveMaximum: 1024
                    exclusiveMinimum: 1000
                    title: Big Int
                    type: integer
                big_float:
                    exclusiveMaximum: 1024.0
                    exclusiveMinimum: 1000.0
                    title: Big Float
                    type: number
                unit_interval:
                    maximum: 1
                    minimum: 0
                    title: Unit Interval
                    type: number
                decimal_positive:
                    anyOf:
                    -   exclusiveMinimum: 0.0
                        type: number
                    -   pattern: ^(?!^[-+.]*$)[+-]?0*\d*\.?\d*$
                        type: string
                    title: Decimal Positive
                decimal_negative:
                    anyOf:
                    -   exclusiveMaximum: 0.0
                        type: number
                    -   pattern: ^(?!^[-+.]*$)[+-]?0*\d*\.?\d*$
                        type: string
                    title: Decimal Negative
            required:
            - a
            - b
            - big_int
            - big_float
            - unit_interval
            - decimal_positive
            - decimal_negative
            title: PydConstrained
            type: object
tags: []
    """.strip()
    assert yaml.strip() == expected_result


async def test_schema_registration(docs: OpenAPIHandler, serializer: Serializer):
    @docs.register(
        Schema(
            type=ValueType.OBJECT,
            required=["foo"],
            properties={
                "foo": Schema(
                    ValueType.INTEGER, minimum=10, maximum=100, nullable=False
                ),
                "ufo": Schema(
                    ValueType.STRING, min_length=5, max_length=10, nullable=False
                ),
            },
        )
    )
    class A: ...

    app = get_app()

    @app.router.route("/")
    def home() -> A: ...

    @docs(
        security=[
            SecurityInfo("basicAuth", []),
            SecurityInfo("bearerAuth", ["read:home", "write:home"]),
        ]
    )
    @app.router.route("/", methods=["POST"])
    def auth_home() -> A: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/A'
            operationId: home
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/A'
            operationId: auth_home
            security:
            -   basicAuth: []
            -   bearerAuth:
                - read:home
                - write:home
components:
    schemas:
        A:
            type: object
            required:
            - foo
            properties:
                foo:
                    type: integer
                    maximum: 100
                    minimum: 10
                    nullable: false
                ufo:
                    type: string
                    maxLength: 10
                    minLength: 5
                    nullable: false
tags: []
""".strip()
    )


async def test_handles_ref_for_optional_type(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.router.route("/cats")
    def one() -> PaginatedSet[Cat]: ...

    @app.router.route("/cats/{cat_id}")
    def two(cat_id: int) -> Cat | None: ...

    @app.router.route("/cats_alt/{cat_id}")
    def three(cat_id: int) -> Cat: ...

    @app.router.route("/cats_value_pattern/{uuid:cat_id}")
    def four(cat_id: UUID) -> Cat: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /cats:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfCat'
            operationId: one
    /cats/{cat_id}:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Cat'
                '404':
                    description: Object not found
            operationId: two
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
    /cats_alt/{cat_id}:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Cat'
            operationId: three
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
    /cats_value_pattern/{cat_id}:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Cat'
            operationId: four
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: string
                    format: uuid
                    nullable: false
                description: ''
                required: true
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Cat'
                total:
                    type: integer
                    format: int64
                    nullable: false
tags: []
""".strip()
    )


async def test_handles_from_form_docs(docs: OpenAPIHandler, serializer: Serializer):
    app = get_app()

    @app.router.post("/foo")
    def one(data: FromForm[CreateFooInput]) -> Foo: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /foo:
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Foo'
            operationId: one
            parameters: []
            requestBody:
                content:
                    multipart/form-data:
                        schema:
                            $ref: '#/components/schemas/CreateFooInput'
                    application/x-www-form-urlencoded:
                        schema:
                            $ref: '#/components/schemas/CreateFooInput'
                required: true
components:
    schemas:
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        CreateFooInput:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
tags: []
""".strip()
    )


async def test_websockets_routes_are_ignored(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.router.post("/foo")
    def one(data: FromForm[CreateFooInput]) -> Foo: ...

    @app.router.ws("/ws")
    def websocket_route() -> None: ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /foo:
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Foo'
            operationId: one
            parameters: []
            requestBody:
                content:
                    multipart/form-data:
                        schema:
                            $ref: '#/components/schemas/CreateFooInput'
                    application/x-www-form-urlencoded:
                        schema:
                            $ref: '#/components/schemas/CreateFooInput'
                required: true
components:
    schemas:
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        CreateFooInput:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
tags: []
""".strip()
    )


async def test_mount_oad_generation(serializer: Serializer):
    """
    Tests support for OAD generation of mounted apps, using the options:

    parent.mount_registry.auto_events = True
    parent.mount_registry.handle_docs = True
    """
    parent = Application(show_error_details=True)
    parent.mount_registry.auto_events = True
    parent.mount_registry.handle_docs = True

    docs = OpenAPIHandler(info=Info(title="Parent API", version="0.0.1"))
    docs.bind_app(parent)

    @dataclass
    class CreateCatInput:
        name: str
        email: str
        foo: int

    @dataclass
    class CreateDogInput:
        name: str
        email: str
        example: int

    @dataclass
    class CreateParrotInput:
        name: str
        email: str

    @parent.router.get("/")
    def a_home():
        """Parent root."""
        return "Hello, from the parent app - for information, navigate to /docs"

    @parent.router.get("/cats")
    def get_cats_conflicting():
        """Conflict! This will be overridden by the child app route!"""
        return "CONFLICT"

    child_1 = Application()

    @child_1.router.get("/")
    def get_cats():
        """Gets a list of cats."""
        return "Gets a list of cats."

    @child_1.router.post("/")
    def create_cat(data: CreateCatInput):
        """Creates a new cat."""
        return "Creates a new cat."

    @child_1.router.delete("/{cat_id}")
    def delete_cat(cat_id: str):
        """Deletes a cat by id."""
        return "Deletes a cat by id."

    child_2 = Application()

    @child_2.router.get("/")
    def get_dogs():
        """Gets a list of dogs."""
        return "Gets a list of dogs."

    @child_2.router.post("/")
    def create_dog(data: CreateDogInput):
        """Creates a new dog."""
        return "Creates a new dog."

    @child_2.router.delete("/{dog_id}")
    def delete_dog(dog_id: str):
        """Deletes a dog by id."""
        return "Deletes a dog by id."

    child_3 = Application()

    @child_3.router.get("/")
    def get_parrots():
        """Gets a list of parrots."""
        return "Gets a list of parrots"

    @child_3.router.post("/")
    def create_parrot(data: CreateParrotInput):
        """Creates a new parrot."""
        return "Creates a new parrot"

    @child_3.router.delete("/{parrot_id}")
    def delete_parrot(parrot_id: str):
        """Deletes a parrot by id."""
        return "Deletes a parrot by id."

    parent.mount("/cats", child_1)
    parent.mount("/dogs", child_2)
    parent.mount("/parrots", child_3)

    await parent.start()

    yaml = serializer.to_yaml(docs.generate_documentation(parent))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Parent API
    version: 0.0.1
paths:
    /:
        get:
            responses: {}
            summary: Parent root.
            description: Parent root.
            operationId: a_home
    /cats:
        get:
            responses: {}
            summary: Gets a list of cats.
            description: Gets a list of cats.
            operationId: get_cats
        post:
            responses: {}
            summary: Creates a new cat.
            description: Creates a new cat.
            operationId: create_cat
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/CreateCatInput'
                required: true
    /cats/{cat_id}:
        delete:
            responses: {}
            summary: Deletes a cat by id.
            description: Deletes a cat by id.
            operationId: delete_cat
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: string
                    nullable: false
                description: ''
                required: true
    /dogs:
        get:
            responses: {}
            summary: Gets a list of dogs.
            description: Gets a list of dogs.
            operationId: get_dogs
        post:
            responses: {}
            summary: Creates a new dog.
            description: Creates a new dog.
            operationId: create_dog
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/CreateDogInput'
                required: true
    /dogs/{dog_id}:
        delete:
            responses: {}
            summary: Deletes a dog by id.
            description: Deletes a dog by id.
            operationId: delete_dog
            parameters:
            -   name: dog_id
                in: path
                schema:
                    type: string
                    nullable: false
                description: ''
                required: true
    /parrots:
        get:
            responses: {}
            summary: Gets a list of parrots.
            description: Gets a list of parrots.
            operationId: get_parrots
        post:
            responses: {}
            summary: Creates a new parrot.
            description: Creates a new parrot.
            operationId: create_parrot
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/CreateParrotInput'
                required: true
    /parrots/{parrot_id}:
        delete:
            responses: {}
            summary: Deletes a parrot by id.
            description: Deletes a parrot by id.
            operationId: delete_parrot
            parameters:
            -   name: parrot_id
                in: path
                schema:
                    type: string
                    nullable: false
                description: ''
                required: true
components:
    schemas:
        CreateCatInput:
            type: object
            required:
            - name
            - email
            - foo
            properties:
                name:
                    type: string
                    nullable: false
                email:
                    type: string
                    nullable: false
                foo:
                    type: integer
                    format: int64
                    nullable: false
        CreateDogInput:
            type: object
            required:
            - name
            - email
            - example
            properties:
                name:
                    type: string
                    nullable: false
                email:
                    type: string
                    nullable: false
                example:
                    type: integer
                    format: int64
                    nullable: false
        CreateParrotInput:
            type: object
            required:
            - name
            - email
            properties:
                name:
                    type: string
                    nullable: false
                email:
                    type: string
                    nullable: false
tags: []
""".strip()
    )


async def test_mount_oad_generation_sub_children(serializer: Serializer):
    """
    Tests support for OAD generation of mounted apps, using the options:

    parent.mount_registry.auto_events = True
    parent.mount_registry.handle_docs = True
    """
    parent = Application(show_error_details=True)
    parent.mount_registry.auto_events = True
    parent.mount_registry.handle_docs = True

    docs = OpenAPIHandler(
        info=Info(title="Parent API", version="0.0.1"),
        tags=[Tag(name="A Home")],
    )
    docs.bind_app(parent)

    @docs(tags=["A Home"])
    @parent.router.get("/")
    def a_home():
        """Parent root."""
        return "Hello, from the parent app - for information, navigate to /docs"

    child_1 = Application()
    child_2 = Application()
    child_3 = Application()

    @docs(tags=["Child z Home"])
    @child_1.router.get("/")
    def child_1_home():
        return "Child 1 home."

    @docs(tags=["Child y Home"])
    @child_2.router.get("/")
    def child_2_home():
        return "Child 2 home."

    @docs(tags=["Child x Home"])
    @child_3.router.get("/")
    def child_3_home():
        return "Child 3 home."

    child_1.mount_registry.auto_events = True
    child_1.mount_registry.handle_docs = True
    child_2.mount_registry.auto_events = True
    child_2.mount_registry.handle_docs = True

    parent.mount("/child-1", child_1)
    child_1.mount("/child-2", child_2)
    child_2.mount("/child-3", child_3)

    await parent.start()

    yaml = serializer.to_yaml(docs.generate_documentation(parent))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Parent API
    version: 0.0.1
paths:
    /:
        get:
            responses: {}
            tags:
            - A Home
            summary: Parent root.
            description: Parent root.
            operationId: a_home
    /child-1:
        get:
            responses: {}
            tags:
            - Child z Home
            operationId: child_1_home
    /child-1/child-2:
        get:
            responses: {}
            tags:
            - Child y Home
            operationId: child_2_home
    /child-1/child-2/child-3:
        get:
            responses: {}
            tags:
            - Child x Home
            operationId: child_3_home
components: {}
tags:
-   name: A Home
""".strip()
    )


async def test_sorting_api_controllers_tags(serializer: Serializer):
    app = get_app()
    get = app.controllers_router.get
    post = app.controllers_router.post

    docs = OpenAPIHandler(
        info=Info(
            title="Example API",
            version="0.0.1",
        ),
        security_schemes={
            "basicAuth": HTTPSecurity(
                scheme="basic",
                description="Basic Auth",
            ),
            "bearerAuth": HTTPSecurity(
                scheme="bearer",
                description="Bearer Auth",
            ),
            "apiKeyAuth": APIKeySecurity(
                in_=ParameterLocation.HEADER,
                name="X-API-Key",
                description="API Key Auth",
            ),
            "openID": OpenIdConnectSecurity(
                open_id_connect_url="https://example.com",
                description="OIDC Auth",
            ),
            "oauth2": OAuth2Security(
                flows=OAuthFlows(
                    implicit=OAuthFlow(
                        authorization_url="https://example.com/oauth2/authorize",
                        token_url="https://example.com/oauth2/token",
                        refresh_url="https://example.com/oauth2/refresh",
                        scopes={
                            "read:cats": "Read your cats",
                            "write:cats": "Write your cats",
                        },
                    )
                ),
                description="OAuth2 Auth",
            ),
        },
    )
    docs.bind_app(app)

    @dataclass
    class Cat:
        pass

    @dataclass
    class Dog:
        pass

    @dataclass
    class Parrot:
        pass

    class Parrots(APIController):
        @get()
        def get_parrots(self) -> list[Parrot]:
            """Return the list of configured Parrots."""

        @post()
        def create_parrot(self, parrot: Parrot) -> None:
            """Add a Parrot to the system."""

    class Dogs(APIController):
        @get()
        def get_dogs(self) -> list[Dog]:
            """Return the list of configured dogs."""

        @post()
        def create_dog(self, dog: Dog) -> None:
            """Add a Dog to the system."""

    class Cats(APIController):
        @get()
        def get_cats(self) -> list[Cat]:
            """Return the list of configured cats."""

        @post()
        def create_cat(self, cat: Cat) -> None:
            """Add a Cat to the system."""

    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example API
    version: 0.0.1
paths:
    /api/parrots:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                type: array
                                nullable: false
                                items:
                                    $ref: '#/components/schemas/Parrot'
            tags:
            - Parrots
            summary: Return the list of configured Parrots.
            description: Return the list of configured Parrots.
            operationId: get_parrots
            parameters: []
        post:
            responses:
                '204':
                    description: Success response
            tags:
            - Parrots
            summary: Add a Parrot to the system.
            description: Add a Parrot to the system.
            operationId: create_parrot
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/Parrot'
                required: true
    /api/dogs:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                type: array
                                nullable: false
                                items:
                                    $ref: '#/components/schemas/Dog'
            tags:
            - Dogs
            summary: Return the list of configured dogs.
            description: Return the list of configured dogs.
            operationId: get_dogs
            parameters: []
        post:
            responses:
                '204':
                    description: Success response
            tags:
            - Dogs
            summary: Add a Dog to the system.
            description: Add a Dog to the system.
            operationId: create_dog
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/Dog'
                required: true
    /api/cats:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                type: array
                                nullable: false
                                items:
                                    $ref: '#/components/schemas/Cat'
            tags:
            - Cats
            summary: Return the list of configured cats.
            description: Return the list of configured cats.
            operationId: get_cats
            parameters: []
        post:
            responses:
                '204':
                    description: Success response
            tags:
            - Cats
            summary: Add a Cat to the system.
            description: Add a Cat to the system.
            operationId: create_cat
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/Cat'
                required: true
components:
    schemas:
        Parrot:
            type: object
            properties: {}
        Dog:
            type: object
            properties: {}
        Cat:
            type: object
            properties: {}
    securitySchemes:
        basicAuth:
            scheme: basic
            type: http
            description: Basic Auth
        bearerAuth:
            scheme: bearer
            type: http
            description: Bearer Auth
        apiKeyAuth:
            name: X-API-Key
            in: header
            type: apiKey
            description: API Key Auth
        openID:
            openIdConnectUrl: https://example.com
            type: openIdConnect
            description: OIDC Auth
        oauth2:
            flows:
                implicit:
                    scopes:
                        read:cats: Read your cats
                        write:cats: Write your cats
                    authorizationUrl: https://example.com/oauth2/authorize
                    tokenUrl: https://example.com/oauth2/token
                    refreshUrl: https://example.com/oauth2/refresh
            type: oauth2
            description: OAuth2 Auth
tags:
-   name: Cats
-   name: Dogs
-   name: Parrots
""".strip()
    )


@dataclass
class A:
    a_prop: int


@dataclass
class B:
    b_prop: str


@dataclass
class C:
    c_prop: str


@dataclass
class D:
    d_prop: float


@dataclass
class E:
    e_prop: int


@dataclass
class F:
    f_prop: str
    f_prop2: A


@dataclass
class AnyOfTestClass:
    sub_prop: Union[A, B, C]


@dataclass
class AnyOfResponseTestClass:
    data: Union[D, E, F]


class APyd(BaseModel):
    a_prop: int


class BPyd(BaseModel):
    b_prop: str


class CPyd(BaseModel):
    c_prop: str


class DPyd(BaseModel):
    d_prop: float


class EPyd(BaseModel):
    e_prop: int


class FPyd(BaseModel):
    f_prop: str
    f_prop2: APyd


class AnyOfTestClassPyd(BaseModel):
    sub_prop: APyd | BPyd | CPyd


class AnyOfResponseTestClassPyd(BaseModel):
    data: DPyd | EPyd | FPyd


async def test_any_of_dataclasses(docs: OpenAPIHandler, serializer: Serializer):
    app = get_app()
    docs.bind_app(app)

    @app.router.post("/one")
    def one(data: AnyOfTestClass) -> AnyOfResponseTestClass: ...

    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    expected_fragments = [
        """
    /one:
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/AnyOfResponseTestClass'
            operationId: one
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/AnyOfTestClass'
                required: true
        """,
        """
        D:
            type: object
            required:
            - d_prop
            properties:
                d_prop:
                    type: number
                    format: float
                    nullable: false
        """,
        """
        UnionOfDAndEAndF:
            type: object
            anyOf:
            -   $ref: '#/components/schemas/D'
            -   $ref: '#/components/schemas/E'
            -   $ref: '#/components/schemas/F'
        """,
        """
        AnyOfResponseTestClass:
            type: object
            properties:
                data:
                    $ref: '#/components/schemas/UnionOfDAndEAndF'
        """,
        """
        UnionOfAAndBAndC:
            type: object
            anyOf:
            -   $ref: '#/components/schemas/A'
            -   $ref: '#/components/schemas/B'
            -   $ref: '#/components/schemas/C'
        """,
    ]

    for fragment in expected_fragments:
        assert fragment.strip() in yaml


async def test_any_of_pydantic_models(docs: OpenAPIHandler, serializer: Serializer):
    app = get_app()
    docs.bind_app(app)

    @app.router.post("/one")
    def one(data: AnyOfTestClassPyd) -> AnyOfResponseTestClassPyd: ...

    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    expected_fragments = [
        """
openapi: 3.1.0
info:
    title: Example
    version: 0.0.1
paths:
    /one:
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/AnyOfResponseTestClassPyd'
            operationId: one
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/AnyOfTestClassPyd'
                required: true
components:
    schemas:
        APyd:
            properties:
                a_prop:
                    title: A Prop
                    type: integer
            required:
            - a_prop
            title: APyd
            type: object
        DPyd:
            properties:
                d_prop:
                    title: D Prop
                    type: number
            required:
            - d_prop
            title: DPyd
            type: object
        EPyd:
            properties:
                e_prop:
                    title: E Prop
                    type: integer
            required:
            - e_prop
            title: EPyd
            type: object
        FPyd:
            properties:
                f_prop:
                    title: F Prop
                    type: string
                f_prop2:
                    $ref: '#/components/schemas/APyd'
            required:
            - f_prop
            - f_prop2
            title: FPyd
            type: object
        AnyOfResponseTestClassPyd:
            properties:
                data:
                    anyOf:
                    -   $ref: '#/components/schemas/DPyd'
                    -   $ref: '#/components/schemas/EPyd'
                    -   $ref: '#/components/schemas/FPyd'
                    title: Data
            required:
            - data
            title: AnyOfResponseTestClassPyd
            type: object
        BPyd:
            properties:
                b_prop:
                    title: B Prop
                    type: string
            required:
            - b_prop
            title: BPyd
            type: object
        CPyd:
            properties:
                c_prop:
                    title: C Prop
                    type: string
            required:
            - c_prop
            title: CPyd
            type: object
        AnyOfTestClassPyd:
            properties:
                sub_prop:
                    anyOf:
                    -   $ref: '#/components/schemas/APyd'
                    -   $ref: '#/components/schemas/BPyd'
                    -   $ref: '#/components/schemas/CPyd'
                    title: Sub Prop
            required:
            - sub_prop
            title: AnyOfTestClassPyd
            type: object
tags: []
        """.strip(),
    ]

    if PYDANTIC_VERSION == 1:
        expected_fragments = [
            """
    /one:
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/AnyOfResponseTestClassPyd'
            operationId: one
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/AnyOfTestClassPyd'
                required: true
            """,
            """
components:
    schemas:
        DPyd:
            title: DPyd
            type: object
            properties:
                d_prop:
                    title: D Prop
                    type: number
            required:
            - d_prop
        EPyd:
            title: EPyd
            type: object
            properties:
                e_prop:
                    title: E Prop
                    type: integer
            required:
            - e_prop
        APyd:
            title: APyd
            type: object
            properties:
                a_prop:
                    title: A Prop
                    type: integer
            required:
            - a_prop
        FPyd:
            title: FPyd
            type: object
            properties:
                f_prop:
                    title: F Prop
                    type: string
                f_prop2:
                    $ref: '#/components/schemas/APyd'
            required:
            - f_prop
            - f_prop2
        AnyOfResponseTestClassPyd:
            title: AnyOfResponseTestClassPyd
            type: object
            properties:
                data:
                    title: Data
                    anyOf:
                    -   $ref: '#/components/schemas/DPyd'
                    -   $ref: '#/components/schemas/EPyd'
                    -   $ref: '#/components/schemas/FPyd'
            required:
            - data
        BPyd:
            title: BPyd
            type: object
            properties:
                b_prop:
                    title: B Prop
                    type: string
            required:
            - b_prop
        CPyd:
            title: CPyd
            type: object
            properties:
                c_prop:
                    title: C Prop
                    type: string
            required:
            - c_prop
        AnyOfTestClassPyd:
            title: AnyOfTestClassPyd
            type: object
            properties:
                sub_prop:
                    title: Sub Prop
                    anyOf:
                    -   $ref: '#/components/schemas/APyd'
                    -   $ref: '#/components/schemas/BPyd'
                    -   $ref: '#/components/schemas/CPyd'
            required:
            - sub_prop
            """,
        ]

    for fragment in expected_fragments:
        assert fragment.strip() in yaml


@pytest.mark.parametrize(
    "name,programming_result,descriptive_result",
    [
        ("Example[Cat]", "Example_Cat", "ExampleOfCat"),
        ("Union[A, B, C]", "Union_A_B_C", "UnionOfAAndBAndC"),
        ("List[Union[A, B, C]]", "List_Union_A_B_C", "ListOfUnionOfAAndBAndC"),
        ("Dict[str, int]", "Dict_str_int", "DictOfstrAndint"),
        ("dict[str, str]", "dict_str_str", "dictOfstrAndstr"),
        ("A | B | C", "A_Or_B_Or_C", "AOrBOrC"),
        ("list[A | B | C]", "list_A_Or_B_Or_C", "listOfAOrBOrC"),
    ],
)
def test_serializer_naming_styles(name, programming_result, descriptive_result):
    # Programming style
    programming_serializer = DefaultSerializer(programming_names=True)
    assert programming_serializer.get_type_name_for_generic(name) == programming_result

    # Descriptive style (current default)
    descriptive_serializer = DefaultSerializer(programming_names=False)
    assert descriptive_serializer.get_type_name_for_generic(name) == descriptive_result


async def test_tags_decorator(serializer: Serializer):
    app = get_app()
    get = app.controllers_router.get
    post = app.controllers_router.post

    docs = OpenAPIHandler(
        info=Info(
            title="Example API",
            version="0.0.1",
        )
    )
    docs.bind_app(app)

    @dataclass
    class Parrot:
        pass

    @docs.tags("TagExample")
    class Parrots(APIController):
        @get()
        def get_parrots(self) -> list[Parrot]:
            """Return the list of configured Parrots."""
            ...

        @post()
        def create_parrot(self, parrot: Parrot) -> None:
            """Add a Parrot to the system."""

    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.1.0
info:
    title: Example API
    version: 0.0.1
paths:
    /api/parrots:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                type: array
                                nullable: false
                                items:
                                    $ref: '#/components/schemas/Parrot'
            tags:
            - TagExample
            summary: Return the list of configured Parrots.
            description: Return the list of configured Parrots.
            operationId: get_parrots
            parameters: []
        post:
            responses:
                '204':
                    description: Success response
            tags:
            - TagExample
            summary: Add a Parrot to the system.
            description: Add a Parrot to the system.
            operationId: create_parrot
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/Parrot'
                required: true
components:
    schemas:
        Parrot:
            type: object
            properties: {}
tags:
-   name: TagExample
""".strip()
    )



import uuid
from collections.abc import MutableSequence
from inspect import isasyncgenfunction
from urllib.parse import parse_qsl, quote_plus

from blacksheep.settings.json import json_settings

from .exceptions import MessageAborted


class Content:
    def __init__(self, content_type: bytes, data: bytes):
        self.type = content_type
        self.body = data
        self.length = len(data)

    async def read(self):
        return self.body


class StreamedContent(Content):
    def __init__(self, content_type: bytes, data_provider, data_length: int = -1):
        self.type = content_type
        self.body = None
        self.length = data_length
        self.generator = data_provider
        if not isasyncgenfunction(data_provider):
            raise ValueError("Data provider must be an async generator")

    async def read(self):
        value = bytearray()
        async for chunk in self.generator():
            value.extend(chunk)
        self.body = bytes(value)
        self.length = len(self.body)
        return self.body

    async def stream(self):
        async for chunk in self.generator():
            yield chunk

    async def get_parts(self):
        async for chunk in self.generator():
            yield chunk


class ASGIContent(Content):
    def __init__(self, receive):
        self.type = None
        self.body = None
        self.length = -1
        self.receive = receive

    def dispose(self):
        self.receive = None
        self.body = None

    async def stream(self):
        while True:
            message = await self.receive()
            if message.get("type") == "http.disconnect":
                raise MessageAborted()
            yield message.get("body", b"")
            if not message.get("more_body"):
                break
        yield b""

    async def read(self):
        if self.body is not None:
            return self.body
        value = bytearray()
        while True:
            message = await self.receive()
            if message.get("type") == "http.disconnect":
                raise MessageAborted()
            value.extend(message.get("body", b""))
            if not message.get("more_body"):
                break
        self.body = bytes(value)
        self.length = len(self.body)
        return self.body


class TextContent(Content):
    def __init__(self, text: str):
        super().__init__(b"text/plain; charset=utf-8", text.encode("utf8"))


class HTMLContent(Content):
    def __init__(self, html: str):
        super().__init__(b"text/html; charset=utf-8", html.encode("utf8"))


class JSONContent(Content):
    def __init__(self, data, dumps=json_settings.dumps):
        super().__init__(b"application/json", dumps(data).encode("utf8"))


def parse_www_form_urlencoded(content: str) -> dict:
    data = {}
    for key, value in parse_qsl(content):
        if key in data:
            if isinstance(data[key], str):
                data[key] = [data[key], value]
            else:
                data[key].append(value)
        else:
            data[key] = value
    return data


def parse_www_form(content: str) -> dict:
    return parse_www_form_urlencoded(content)


def try_decode(value: bytes, encoding: str):
    try:
        return value.decode(encoding or "utf8")
    except Exception:
        return value


def multiparts_to_dictionary(parts: list) -> dict:
    data = {}
    for part in parts:
        key = part.name.decode("utf8")
        charset = part.charset.encode() if part.charset else None
        if part.file_name:
            if key in data:
                data[key].append(part)
            else:
                data[key] = [part]
        else:
            if key in data:
                if isinstance(data[key], list):
                    data[key].append(try_decode(part.data, charset))
                else:
                    data[key] = [data[key], try_decode(part.data, charset)]
            else:
                data[key] = try_decode(part.data, charset)
    return data


def write_multipart_part(part, destination: bytearray):
    destination.extend(b'Content-Disposition: form-data; name="')
    destination.extend(part.name)
    destination.extend(b'"')
    if part.file_name:
        destination.extend(b'; filename="')
        destination.extend(part.file_name)
        destination.extend(b'"\r\n')
    if part.content_type:
        destination.extend(b"Content-Type: ")
        destination.extend(part.content_type)
    destination.extend(b"\r\n\r\n")
    destination.extend(part.data)
    destination.extend(b"\r\n")


def write_www_form_urlencoded(data: dict | list) -> bytes:
    if isinstance(data, list):
        values = data
    else:
        values = data.items()
    contents = []
    for key, value in values:
        if isinstance(value, MutableSequence):
            for item in value:
                contents.append(quote_plus(key) + "=" + quote_plus(str(item)))
        else:
            contents.append(quote_plus(key) + "=" + quote_plus(str(value)))
    return ("&".join(contents)).encode("utf8")


class FormContent(Content):
    def __init__(self, data: dict[str, str] | list[tuple[str, str]]):
        super().__init__(
            b"application/x-www-form-urlencoded", write_www_form_urlencoded(data)
        )


class FormPart:
    def __init__(
        self,
        name: bytes,
        data: bytes,
        content_type: bytes | None = None,
        file_name: bytes | None = None,
        charset: bytes | None = None,
    ):
        self.name = name
        self.data = data
        self.file_name = file_name
        self.content_type = content_type
        self.charset = charset

    def __eq__(self, other):
        if isinstance(other, FormPart):
            return (
                other.name == self.name
                and other.file_name == self.file_name
                and other.content_type == self.content_type
                and other.charset == self.charset
                and other.data == self.data
            )
        if other is None:
            return False
        return NotImplemented

    def __repr__(self):
        return f"<FormPart {self.name} - at {id(self)}>"


class MultiPartFormData(Content):
    def __init__(self, parts: list):
        self.parts = parts
        self.boundary = b"------" + str(uuid.uuid4()).replace("-", "").encode()
        super().__init__(
            b"multipart/form-data; boundary=" + self.boundary,
            write_multipart_form_data(self),
        )


def write_multipart_form_data(data: "MultiPartFormData") -> bytes:
    contents = bytearray()
    for part in data.parts:
        contents.extend(b"--")
        contents.extend(data.boundary)
        contents.extend(b"\r\n")
        write_multipart_part(part, contents)
    contents.extend(b"--")
    contents.extend(data.boundary)
    contents.extend(b"--\r\n")
    return bytes(contents)


class ServerSentEvent:
    """
    Represents a single event of a Server-sent event communication, to be used
    in a asynchronous generator.
    """

    def __init__(
        self,
        data,
        event: str | None = None,
        id: str | None = None,
        retry: int | None = -1,
        comment: str | None = None,
    ):
        self.data = data
        self.event = event
        self.id = id
        self.retry = retry
        self.comment = comment

    def write_data(self) -> str:
        return json_settings.dumps(self.data)

    def __repr__(self):
        return f"ServerSentEvent({self.data})"


class TextServerSentEvent(ServerSentEvent):
    def __init__(
        self,
        data: str,
        event: str | None = None,
        id: str | None = None,
        retry: int | None = -1,
        comment: str | None = None,
    ):
        super().__init__(data, event, id, retry, comment)

    def write_data(self) -> str:
        return self.data.replace("\r", "\\r").replace("\n", "\\n")



import pytest

from blacksheep.messages import Request
from blacksheep.server.redirects import (
    default_trailing_slash_exclude,
    get_trailing_slash_middleware,
)
from blacksheep.server.responses import text


class TestDefaultTrailingSlashExclude:
    def test_excludes_api_paths(self):
        assert default_trailing_slash_exclude("/api/users") is True
        assert default_trailing_slash_exclude("/api/") is True
        assert default_trailing_slash_exclude("/v1/api/endpoint") is True

    def test_does_not_exclude_non_api_paths(self):
        assert default_trailing_slash_exclude("/home") is False
        assert default_trailing_slash_exclude("/about") is False
        assert default_trailing_slash_exclude("/") is False


class TestTrailingSlashMiddleware:
    @pytest.mark.asyncio
    async def test_redirects_path_without_trailing_slash(self):
        middleware = get_trailing_slash_middleware()
        request = Request("GET", b"/home", [])

        async def handler(req):
            return text("OK")

        response = await middleware(request, handler)

        assert response.status == 301
        assert response.headers.get_single(b"Location") == b"/home/"

    @pytest.mark.asyncio
    async def test_does_not_redirect_path_with_trailing_slash(self):
        middleware = get_trailing_slash_middleware()
        request = Request("GET", b"/home/", [])

        async def handler(req):
            return text("OK")

        response = await middleware(request, handler)

        assert response.status == 200

    @pytest.mark.asyncio
    async def test_does_not_redirect_paths_with_file_extensions(self):
        middleware = get_trailing_slash_middleware()
        request = Request("GET", b"/style.css", [])

        async def handler(req):
            return text("OK")

        response = await middleware(request, handler)

        assert response.status == 200

    @pytest.mark.asyncio
    async def test_does_not_redirect_paths_with_file_extensions_in_subdirs(self):
        middleware = get_trailing_slash_middleware()
        request = Request("GET", b"/assets/script.js", [])

        async def handler(req):
            return text("OK")

        response = await middleware(request, handler)

        assert response.status == 200

    @pytest.mark.asyncio
    async def test_excludes_api_paths_by_default(self):
        middleware = get_trailing_slash_middleware()
        request = Request("GET", b"/api/users", [])

        async def handler(req):
            return text("OK")

        response = await middleware(request, handler)

        assert response.status == 200

    @pytest.mark.asyncio
    async def test_custom_exclude_function(self):
        def custom_exclude(path: str) -> bool:
            return path.startswith("/admin")

        middleware = get_trailing_slash_middleware(exclude=custom_exclude)
        request = Request("GET", b"/admin/dashboard", [])

        async def handler(req):
            return text("OK")

        response = await middleware(request, handler)

        assert response.status == 200

    @pytest.mark.asyncio
    async def test_custom_exclude_does_not_affect_other_paths(self):
        def custom_exclude(path: str) -> bool:
            return path.startswith("/admin")

        middleware = get_trailing_slash_middleware(exclude=custom_exclude)
        request = Request("GET", b"/home", [])

        async def handler(req):
            return text("OK")

        response = await middleware(request, handler)

        assert response.status == 301
        assert response.headers.get_single(b"Location") == b"/home/"

    @pytest.mark.asyncio
    async def test_none_exclude_disables_exclusion(self):
        middleware = get_trailing_slash_middleware(exclude=lambda x: False)
        request = Request("GET", b"/api/users", [])

        async def handler(req):
            return text("OK")

        response = await middleware(request, handler)

        assert response.status == 301
        assert response.headers.get_single(b"Location") == b"/api/users/"

    @pytest.mark.asyncio
    async def test_handles_root_path(self):
        middleware = get_trailing_slash_middleware()
        request = Request("GET", b"/", [])

        async def handler(req):
            return text("OK")

        response = await middleware(request, handler)

        assert response.status == 200

    @pytest.mark.asyncio
    async def test_normalizes_path_with_leading_slashes(self):
        middleware = get_trailing_slash_middleware()
        request = Request("GET", b"/home", [])

        async def handler(req):
            return text("OK")

        response = await middleware(request, handler)

        assert response.status == 301
        assert response.headers.get_single(b"Location") == b"/home/"

    @pytest.mark.asyncio
    async def test_nested_paths_without_trailing_slash(self):
        middleware = get_trailing_slash_middleware()
        request = Request("GET", b"/about/team", [])

        async def handler(req):
            return text("OK")

        response = await middleware(request, handler)

        assert response.status == 301
        assert response.headers.get_single(b"Location") == b"/about/team/"



import http
import inspect
import logging
from collections import UserDict

from blacksheep.server.errors import ServerErrorDetailsHandler
from blacksheep.server.routing import Router

from .contents import Content, TextContent
from .exceptions import HTTPException, InternalServerError, InvalidExceptionHandler
from .messages import Response
from .utils import get_class_instance_hierarchy

try:
    from pydantic import ValidationError
except ImportError:
    ValidationError = None


class ExceptionHandlersDict(UserDict):

    def __setitem__(self, key, item) -> None:
        if not inspect.iscoroutinefunction(item):
            raise InvalidExceptionHandler()
        signature = inspect.Signature.from_callable(item)
        if len(signature.parameters) != 3 and not any(
            param
            for param in signature.parameters
            if signature.parameters[param].kind == 2
        ):
            raise InvalidExceptionHandler()
        return super().__setitem__(key, item)


async def handle_not_found(app, request, http_exception):
    return Response(404, content=TextContent("Resource not found"))


async def handle_internal_server_error(app, request, exception):
    return Response(500, content=TextContent("Internal Server Error"))


async def handle_bad_request(app, request, http_exception):
    if getattr(http_exception, "__context__", None) is not None and callable(
        getattr(http_exception.__context__, "json", None)
    ):
        return Response(
            http_exception.status,
            content=Content(
                b"application/json", http_exception.__context__.json().encode("utf8")
            ),
        )
    return Response(400, content=TextContent(f"Bad Request: {str(http_exception)}"))


async def _default_pydantic_validation_error_handler(app, request, error):
    return Response(
        400, content=Content(b"application/json", error.json(indent=4).encode("utf-8"))
    )


async def common_http_exception_handler(app, request, http_exception):
    return Response(
        http_exception.status,
        content=TextContent(http.HTTPStatus(http_exception.status).phrase),
    )


def get_logger():
    logger = logging.getLogger("blacksheep.server")
    logger.setLevel(logging.INFO)
    return logger


class BaseApplication:
    router: Router

    def __init__(self, show_error_details, router):
        self.router = router
        self.exceptions_handlers = self.init_exceptions_handlers()
        self.show_error_details = show_error_details
        self.logger = get_logger()
        self.server_error_details_handler: ServerErrorDetailsHandler

    def init_exceptions_handlers(self):
        default_handlers = ExceptionHandlersDict(
            {404: handle_not_found, 400: handle_bad_request}
        )
        if ValidationError is not None:
            default_handlers[ValidationError] = (
                _default_pydantic_validation_error_handler
            )
        return default_handlers

    async def log_unhandled_exc(self, request, exc):
        self.logger.error(
            'Unhandled exception - "%s %s"',
            request.method,
            request.url.value.decode(),
            exc_info=exc,
        )

    async def log_handled_exc(self, request, exc):
        if isinstance(exc, HTTPException):
            self.logger.info(
                'HTTP %s - "%s %s". %s',
                exc.status,
                request.method,
                request.url.value.decode(),
                str(exc),
            )
        else:
            self.logger.info(
                'Handled error: "%s %s". %s',
                request.method,
                request.url.value.decode(),
                str(exc),
            )

    async def handle(self, request):
        route = self.router.get_match(request)

        if not route:
            # This is intentional. We should not use user-defined not found exception
            # handlers here because middlewares are not executed. The main router should
            # always have a fallback route configured.
            return Response(404)

        request.route_values = route.values
        try:
            response = await route.handler(request)
        except Exception as exc:
            response = await self.handle_request_handler_exception(request, exc)
        return response or Response(204)

    async def handle_request_handler_exception(self, request, exc):
        if isinstance(exc, HTTPException):
            await self.log_handled_exc(request, exc)
            return await self.handle_http_exception(request, exc)
        if self.is_handled_exception(exc):
            await self.log_handled_exc(request, exc)
        else:
            await self.log_unhandled_exc(request, exc)
        return await self.handle_exception(request, exc)

    def get_http_exception_handler(self, http_exception):
        handler = self.get_exception_handler(http_exception, stop_at=HTTPException)
        if handler:
            return handler
        return self.exceptions_handlers.get(
            getattr(http_exception, "status", None), common_http_exception_handler
        )

    def is_handled_exception(self, exception):
        for class_type in get_class_instance_hierarchy(exception):
            if class_type in self.exceptions_handlers:
                return True
        return False

    def get_exception_handler(self, exception, stop_at):
        for class_type in get_class_instance_hierarchy(exception):
            if stop_at is not None and stop_at is class_type:
                return None
            if class_type in self.exceptions_handlers:
                return self.exceptions_handlers[class_type]
        return None

    async def handle_internal_server_error(self, request, exc):
        if self.show_error_details:
            return self.server_error_details_handler.produce_response(request, exc)
        error = InternalServerError(exc)
        internal_server_error_handler = self.get_http_exception_handler(error)
        try:
            return await internal_server_error_handler(self, request, error)
        except Exception:
            self.logger.exception(
                "An exception occurred while trying to apply the configured "
                "Internal Server Error handler!"
            )
        return Response(500, content=TextContent("Internal Server Error"))

    async def _apply_exception_handler(self, request, exc, exception_handler):
        try:
            return await exception_handler(self, request, exc)
        except Exception as server_ex:
            # If the exception happens in the user-defined exception handler,
            # we need to fallback to the default handlers.
            self.logger.error(
                "Unhandled exception in exception_handler: %s",
                exception_handler.__name__,
            )
            if self.show_error_details:
                return self.server_error_details_handler.produce_response(request, exc)

            return await handle_internal_server_error(self, request, server_ex)

    async def handle_http_exception(self, request, http_exception):
        exception_handler = self.get_http_exception_handler(http_exception)
        if exception_handler:
            return await self._apply_exception_handler(
                request, http_exception, exception_handler
            )
        return await self.handle_exception(request, http_exception)

    async def handle_exception(self, request, exc):
        exception_handler = self.get_exception_handler(exc, None)
        if exception_handler:
            return await self._apply_exception_handler(request, exc, exception_handler)
        return await self.handle_internal_server_error(request, exc)
